"""
********************************************************************************
compas_xr.realtime_database
********************************************************************************

This package contains classes for using Firebase realtime database.

.. currentmodule:: compas_xr.realtime_database

Classes
-------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    RealtimeDatabase

"""

import sys

if sys.platform == "cli":
    from compas_xr.realtime_database.realtime_database_cli import RealtimeDatabase
else:
    from compas_xr.realtime_database.realtime_database_pyrebase import RealtimeDatabase

__all__ = ["RealtimeDatabase"]



"""
Sync Results.

The Sync Result component is used to consolidate all user-defined inputs for the resulting trajectory and coordinate them into a single, unified trajectory.

COMPAS XR v0.9.1
"""

from ghpythonlib.componentbase import executingcomponent as component

from compas_xr.ghpython import TrajectoryResultManager


class SyncResultComponent(component):
    def RunScript(self, element_id, trajectory, robot_base_frame, pick_and_place, pick_index, ee_link_name, options):
        if element_id:
            result = TrajectoryResultManager()
            result.requested_element_id = element_id
            result.robot_base_frame = robot_base_frame
            result.trajectory = result.format_trajectory(trajectory)
            if pick_and_place is not None:
                result.pick_and_place = pick_and_place
                if pick_and_place:
                    result.pick_index = pick_index
                    result.end_effector_link_name = ee_link_name
            else:
                result.pick_and_place = False
        else:
            result = None



import sys
import threading
import uuid
from datetime import datetime

from compas.geometry import Frame
from compas_eve import Message


class SequenceCounter(object):
    """An atomic, thread-safe sequence increament counter that increments with each message."""

    ROLLOVER_THRESHOLD = sys.maxsize

    def __init__(self, start=0):
        """Initialize a new counter to given initial value."""
        self._lock = threading.Lock()
        self._value = start

    def increment(self, num=1):
        """Atomically increment the counter by ``num`` and
        return the new value.
        """
        with self._lock:
            self._value += num
            if self._value > SequenceCounter.ROLLOVER_THRESHOLD:
                self._value = 1
            return self._value

    def update_from_msg(self, value):
        """Method to compare value received and current value
        if it is greater then current value update my current value.
        """
        with self._lock:
            if value > self._value:
                self._value = value


class ResponseID(object):
    """An atomic, thread-safe counter that increments with each service routine."""

    ROLLOVER_THRESHOLD = sys.maxsize

    def __init__(self, start=0):
        """Initialize a new counter to given initial value."""
        self._lock = threading.Lock()
        self._value = start

    def increment(self, num=1):
        """Atomically increment the counter by ``num`` and
        return the new value.
        """
        with self._lock:
            self._value += num
            if self._value > ResponseID.ROLLOVER_THRESHOLD:
                self._value = 1
            return self._value

    def update_from_msg(self, value):
        """Method to compare value received and current value
        if it is greater then current value update my current value.
        """
        with self._lock:
            if value > self._value:
                self._value = value


class Header(Message):
    """
    The header class is responsible for coordinating and understanding messages between users.

    The Header class provides methods for parsing, updating, and accessing the header fields of a message,
    and provides a means of defining attributes of the message in order to accept or ignore specific messages.

    Parameters
    ----------
    increment_response_ID : bool, optional
        Whether to increment the response ID when creating a new instance of Header.
    sequence_id : int, optional
        The sequence ID of the message. Optional for parsing.
    response_id : int, optional
        The response ID of the message. Optional for parsing.
    device_id : str, optional
        The device ID of the message. Optional for parsing.
    time_stamp : str, optional
        The timestamp of the message. Optional for parsing.

    Attributes
    ----------
    increment_response_ID : bool
        Whether to increment the response ID when creating a new instance of Header.
    sequence_id : int
        Sequence ID is an atomic counter that increments with each message.
    response_id : int
        Response ID is an int that increments with request routine.
    device_id : str
        Device ID coresponds to the unique system identifier that send the message.
    time_stamp : str
        Timestamp is the time in which the message was sent.
    """

    _shared_sequence_counter = None
    _shared_response_id_counter = None
    _device_id = None

    def __init__(self, increment_response_ID=False, sequence_id=None, response_id=None, device_id=None, time_stamp=None):
        super(Header, self).__init__()
        self["sequence_id"] = sequence_id or self._ensure_sequence_id()
        self["response_id"] = response_id or self._ensure_response_id(increment_response_ID)
        self["device_id"] = device_id or self._get_device_id()
        self["time_stamp"] = time_stamp or self._get_time_stamp()

    @classmethod
    def parse(cls, value):
        """Parse the header information
        from the input value
        """
        instance = cls(value["sequence_id"], value["response_id"], value["device_id"], value["time_stamp"])
        instance._update_sequence_counter_from_message(value["sequence_id"])
        instance._update_response_id_from_message(value["response_id"])
        return instance

    def _get_device_id(self):
        """Ensure device ID is set and return it.
        If not set, generate a new device ID.
        """
        if not Header._device_id:
            Header._device_id = str(uuid.uuid4())
            self.device_id = Header._device_id
        else:
            self.device_id = Header._device_id
        return self.device_id

    def _get_time_stamp(self):
        """Generate timestamp and return it."""
        self.time_stamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")
        return self.time_stamp

    def _ensure_sequence_id(self):
        """Ensure SequenceID is set and return it.
        If not set, generate new shared counter.
        """
        if not Header._shared_sequence_counter:
            Header._shared_sequence_counter = SequenceCounter()
            self.sequence_id = Header._shared_sequence_counter._value
        else:
            self.sequence_id = Header._shared_sequence_counter.increment()
        return self.sequence_id

    def _ensure_response_id(self, increment_response_ID=False):
        """Ensure ResponseID is set and return it.
        If not set, generate new shared counter.
        """
        if not Header._shared_response_id_counter:
            Header._shared_response_id_counter = ResponseID()
            self.response_id = Header._shared_response_id_counter._value
        else:
            if increment_response_ID:
                self.response_id = Header._shared_response_id_counter.increment()
                self.response_id = Header._shared_response_id_counter._value
            else:
                self.response_id = Header._shared_response_id_counter._value
        return self.response_id

    def _update_sequence_counter_from_message(self, sequence_id):
        """Update SequnceID Value if the message input is greater then
        current value +1. Used to ensure messages across devices are in sync
        upon receiving a message if device is restarted.
        """
        if Header._shared_sequence_counter is not None:
            Header._shared_sequence_counter.update_from_msg(sequence_id)
        else:
            Header._shared_sequence_counter = SequenceCounter(start=sequence_id)

    def _update_response_id_from_message(self, response_id):
        """Update ResponseID Value if the message input is greater then
        current value +1. Used to ensure messages across devices are in sync
        upon receiving a message if device is restarted.
        """
        if Header._shared_response_id_counter is not None:
            Header._shared_response_id_counter.update_from_msg(response_id)
        else:
            Header._shared_response_id_counter = ResponseID(start=response_id)

    def update_ids_from_message(self, sequence_id, response_id):
        """Update SequenceID and ResponseID values based on message inputs."""
        self._update_sequence_counter_from_message(sequence_id)
        self._update_response_id_from_message(response_id)


class GetTrajectoryRequest(Message):
    """
    The GetTrajectoryRequest class represents a request message from a user
    to the CAD for retrieving a trajectory.

    Parameters
    ----------
    element_id : str
        The ID of the element associated with the trajectory.
    robot_name : str
        The name of the robot associated with the trajectory.
    header : Header, optional
        The header object containing additional message information.

    Attributes
    ----------
    header : Header
        The header object containing additional message information.
    element_id : str
        The ID of the step in the BuildingPlan associated with the trajectory.
    robot_name : str
        The name of the robot associated with the trajectory.
    trajectory_id : str
        The ID of the trajectory. Default is ``"trajectory_id_" + str(element_id)``.
    """

    def __init__(self, element_id, robot_name, header=None, *args, **kwargs):
        super(GetTrajectoryRequest, self).__init__(*args, **kwargs)
        self["header"] = header or Header(increment_response_ID=True)
        self["element_id"] = element_id
        self["robot_name"] = robot_name
        self["trajectory_id"] = "trajectory_id_" + str(element_id)

    @classmethod
    def parse(cls, data):
        """Parse the GetTrajectoryRequest message from the input data.
        Starts by parsing the header information and then the Message.
        """
        return cls(data["element_id"], data["robot_name"], Header.parse(data["header"]))


class GetTrajectoryResult(Message):
    """
    The GetTrajectoryResult class represents a response message from the CAD
    to all active devices containing a retrieved trajectory.

    Parameters
    ----------
    element_id : str
        The ID of the element associated with the trajectory.
    robot_name : str
        The name of the robot associated with the trajectory.
    robot_base_frame : compas.geometry.Frame
        The base frame of the robot.
    trajectory : dict of joint names and joint values
        The retrieved trajectory.
    header : Header, optional
        The header object containing additional message information.

    Attributes
    ----------
    header : Header
        The header object containing additional message information.
    element_id : str
        The ID of the step in the BuildingPlan associated with the trajectory.
    robot_name : str
        The name of the robot associated with the trajectory.
    robot_base_frame : compas.geometry.Frame
        The base frame of the robot.
    trajectory_id : str
        The ID of the trajectory. Default is ``"trajectory_id_" + str(element_id)``.
    trajectory : dict of joint names and joint values
        The trajectory information computed for the request.
    """

    def __init__(self, element_id, robot_name, robot_base_frame, trajectory, pick_and_place=False, pick_index=None, end_effector_link_name=None, header=None):
        super(GetTrajectoryResult, self).__init__()
        self["header"] = header or Header()
        self["element_id"] = element_id
        self["robot_name"] = robot_name
        self["robot_base_frame"] = robot_base_frame
        self["trajectory_id"] = "trajectory_id_" + str(element_id)
        self["pick_and_place"] = pick_and_place
        self["pick_index"] = pick_index
        self["end_effector_link_name"] = end_effector_link_name
        self["trajectory"] = trajectory

    @classmethod
    def parse(cls, data):
        """Parse the GetTrajectoryResult message from the input data.
        Starts by parsing the header information and then the Message.
        """
        return cls(
            data["element_id"],
            data["robot_name"],
            Frame(**data["robot_base_frame"]),
            data["trajectory"],
            data["pick_and_place"],
            data["pick_index"],
            data["end_effector_link_name"],
            Header.parse(data["header"]),
        )


class ApproveTrajectory(Message):
    """
    The ApproveTrajectory class represents a response message between
    all active devices containing an approval decision for each user.

    Parameters
    ----------
    element_id : str
        The ID of the element associated with the trajectory.
    robot_name : str
        The name of the robot associated with the trajectory.
    trajectory : dict of joint names and joint values
        The approved trajectory.
    approval_status : int
        The approval status of the trajectory.
    header : Header, optional
        The header object containing additional message information.

    Attributes
    ----------
    header : Header
        The header object containing additional message information.
    element_id : str
        The ID of the step in the BuildingPlan associated with the trajectory.
    robot_name : str
        The name of the robot associated with the trajectory.
    trajectory_id : str
        The ID of the trajectory. Default is ``"trajectory_id_" + str(element_id)``.
    trajectory : dict of joint names and joint values
        The approved trajectory.
    approval_status : int
        The approval status of the trajectory.
        0: Not Approved, 1: Approved, 2: Consensus Approval, 3: Cancelation.
    """

    def __init__(self, element_id, robot_name, trajectory, approval_status, header=None):
        super(ApproveTrajectory, self).__init__()
        self["header"] = header or Header()
        self["element_id"] = element_id
        self["robot_name"] = robot_name
        self["trajectory_id"] = "trajectory_id_" + str(element_id)
        self["trajectory"] = trajectory
        self["approval_status"] = approval_status

    @classmethod
    def parse(cls, data):
        """Parse the ApproveTrajectory message from the input data.
        Starts by parsing the header information and then the Message.
        """
        return cls(
            data["element_id"],
            data["robot_name"],
            data["trajectory"],
            data["approval_status"],
            Header.parse(data["header"]),
        )


class ApprovalCounterRequest(Message):
    """
    The ApprovalCounterRequest class represents a request message from a single user
    to all active users to retrieve a count of all activie devices.

    Parameters
    ----------
    element_id : str
        The ID of the element associated with the approval counter.
    header : Header, optional
        The header object containing additional message information.

    Attributes
    ----------
    header : Header
        The header object containing additional message information.
    element_id : str
        The ID of the element associated with the approval counter.
    trajectory_id : str
        The ID of the trajectory. Default is ``"trajectory_id_" + str(element_id)``.
    """

    def __init__(self, element_id, header=None):
        super(ApprovalCounterRequest, self).__init__()
        self["header"] = header or Header()
        self["element_id"] = element_id
        self["trajectory_id"] = "trajectory_id_" + str(element_id)

    @classmethod
    def parse(cls, data):
        """Construct an object of this type from the provided data to support COMPAS JSON serialization.

        Parameters
        ----------
        data : dict
            The raw Python data representing the object.

        Returns
        -------
        object
        """
        return cls(data["element_id"], Header.parse(data["header"]))


class ApprovalCounterResult(Message):
    """
    The ApprovalCounterResult class represents a response message from all active devices
    containing to notify the primary device of the users listening.

    Parameters
    ----------
    element_id : str
        The ID of the element associated with the approval counter.
    header : Header, optional
        The header object containing additional message information.

    Attributes
    ----------
    header : Header
        The header object containing additional message information.
    element_id : str
        The ID of the element associated with the approval counter.
    trajectory_id : str
        The ID of the trajectory. Default is ``"trajectory_id_" + str(element_id)``.
    """

    def __init__(self, element_id, header=None):
        super(ApprovalCounterResult, self).__init__()
        self["header"] = header or Header()
        self["element_id"] = element_id
        self["trajectory_id"] = "trajectory_id_" + str(element_id)

    @classmethod
    def parse(cls, data):
        """Parse the ApprovalCounterResult message from the input data.
        Starts by parsing the header information and then the Message.
        """
        return cls(data["element_id"], Header.parse(data["header"]))


class SendTrajectory(Message):
    """
    The SendTrajectory class represents a message from a user to the CAD
    to give the Approval for Robotic Exacution.

    Parameters
    ----------
    element_id : str
        The ID of the element associated with the trajectory.
    robot_name : str
        The name of the robot associated with the trajectory.
    trajectory : dict of joint names and joint values
        The trajectory to be sent.
    header : Header, optional
        The header object containing additional message information.

    Attributes
    ----------
    header : Header
        The header object containing additional message information.
    element_id : str
        The ID of the element associated with the trajectory.
    robot_name : str
        The name of the robot associated with the trajectory.
    trajectory_id : str
        The ID of the trajectory. Default is ``"trajectory_id_" + str(element_id)``.
    trajectory : dict of joint names and joint values
        The trajectory to be sent.
    """

    def __init__(self, element_id, robot_name, trajectory, header=None):
        super(SendTrajectory, self).__init__()
        self["header"] = header or Header()
        self["element_id"] = element_id
        self["robot_name"] = robot_name
        self["trajectory_id"] = "trajectory_id_" + str(element_id)
        self["trajectory"] = trajectory

    @classmethod
    def parse(cls, data):
        """Parse the SendTrajectory message from the input data.
        Starts by parsing the header information and then the Message.
        """
        return cls(data["element_id"], data["robot_name"], data["trajectory"], Header.parse(data["header"]))



"""
********************************************************************************
compas_xr.mqtt
********************************************************************************

This package contains classes for interfacing with MQTT protocol.

.. currentmodule:: compas_xr.mqtt

MQTT Messages
-------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    GetTrajectoryRequest
    GetTrajectoryResult
    ApproveTrajectory
    SendTrajectory

"""

from .messages import ApproveTrajectory, GetTrajectoryRequest, GetTrajectoryResult, SendTrajectory

__all__ = ["GetTrajectoryRequest", "GetTrajectoryResult", "ApproveTrajectory", "SendTrajectory"]



"""
********************************************************************************
compas_xr
********************************************************************************

.. currentmodule:: compas_xr


.. toctree::
    :maxdepth: 1

    compas_xr.ghpython
    compas_xr.mqtt
    compas_xr.project
    compas_xr.realtime_database
    compas_xr.storage

"""

from __future__ import print_function

import os

__author__ = ["Joseph Kenny"]
__copyright__ = "ETH Zurich, Princeton University"
__license__ = "MIT License"
__email__ = "kenny@arch.ethz.ch"
__version__ = "0.9.1"


HERE = os.path.dirname(__file__)
DATA = os.path.abspath(os.path.join(HERE, "data"))

__all_plugins__ = ["compas_xr.rhino.install"]
__all__ = ["HERE", "DATA"]



from compas_timber.planning import BuildingPlan
from compas_timber.planning import SimpleSequenceGenerator
from compas_timber.planning import Step


class BuildingPlanExtensions(object):
    """
    BuildingPlanExtensions is a class for extending the functionality of the :class:`~compas_timber.planning.BuildingPlan` class.

    The BuildingPlanExtensions class provides additional functionalities to the :class:`~compas_timber.planning.BuildingPlan`
    by providing a way to create a buildling plan from an established assembly sequence.
    """

    # TODO: This makes the building plan in a very manual way
    # TODO: but this needs to be resolved in tandem with building plan revisions.
    def create_buildingplan_from_assembly_sequence(self, assembly, data_type, robot_keys, priority_keys_lists):
        """
        Create a compas_timber.planning.BuildingPlan based on the sequence of the assembly parts.

        Parameters
        ----------
        assembly : :class:`~compas_timber.assembly.TimberAssembly` or :class:`~compas.datastructures.Assembly`
            The assembly that you want to generate the buiding plan for.
        data_type : int
            List index of which data type will be loaded on the application side [0: 'Cylinder', 1: 'Box', 2: 'ObjFile']
        robot_keys : list of str
            List of keys that are intended to be built by the robot.
        priority_keys_lists : list of list of str
            List in assembly order of lists of assembly keys that can be built in parallel.

        Returns
        -------
        building_plan : :class:`~compas_timber.planning.BuildingPlan`
            The building plan generated from the assembly sequence.

        """
        data_type_list = ["0.Cylinder", "1.Box", "2.ObjFile"]
        building_plan = SimpleSequenceGenerator(assembly=assembly).result

        for step in building_plan.steps:
            step.geometry = data_type_list[data_type]
            # TODO: These are unused for now, but are expeted on the application side
            step.instructions = ["none"]
            step.elements_held = [0]

            element_key = str(step.element_ids[0])
            if robot_keys:
                if element_key in robot_keys:
                    step.actor = "ROBOT"

            if not priority_keys_lists:
                step.priority = 0
            else:
                for i, keys_list in enumerate(priority_keys_lists):
                    if element_key in keys_list:
                        step.priority = i
                        break

        return building_plan

    def create_buildingplan_from_with_custom_sequence(self, assembly, sequenced_keys, data_type, robot_keys, priority_keys_lists):
        """
        Create a compas_timber.planning.BuildingPlan based on the sequence of the assembly parts.

        Parameters
        ----------
        assembly : compas_timber.assembly.TimberAssembly or compas.datastructures.Assembly
            The assembly that you want to generate the buiding plan for.
        sequenced_keys : list of str
            List of keys that are intended to be built in the order provided.
        data_type : int
            List index of which data type will be loaded on the application side [0: 'Cylinder', 1: 'Box', 2: 'ObjFile']
        robot_keys : list of str
            List of keys that are intended to be built by the robot.
        priority_keys_lists : list of list of str
            List in assembly order of lists of assembly keys that can be built in parallel.

        Returns
        -------
        building_plan : compas_timber.planning.BuildingPlan
            The building plan generated from the assembly sequence.

        """
        data_type_list = ["0.Cylinder", "1.Box", "2.ObjFile"]
        graph_data = assembly.graph.__data__
        node_data = graph_data["node"]
        building_plan = BuildingPlan()

        for key in sequenced_keys:
            step = Step(key)
            # TODO: This is dumb, but the element_ids are generated incorrectly so they are overwritten here
            step.element_ids = [key]
            step.geometry = data_type_list[data_type]
            # TODO: These are unused for now, but are expeted on the application side
            step.instructions = ["none"]
            step.elements_held = [0]
            step.location = node_data[str(key)]["part"].frame

            if robot_keys:
                if key in robot_keys:
                    step.actor = "ROBOT"
                else:
                    step.actor = "HUMAN"
            else:
                step.actor = "HUMAN"

            if not priority_keys_lists:
                step.priority = 0
            else:
                for i, keys_list in enumerate(priority_keys_lists):
                    if key in keys_list:
                        step.priority = i
                        break
            building_plan.add_step(step)

        return building_plan



class MqttMessageOptionsXR(object):
    def __init__(self, host, project_name, robot_name):
        self.host = host
        self.project_name = project_name
        self.robot_name = robot_name

    def ToString(self):
        return str(self)

    def __str__(self):
        return "Options, host={}, project_name={}, robot_name={}".format(self.host, self.project_name, self.robot_name)



class TrajectoryResultManager(object):
    trajectory = None
    requested_element_id = None
    robot_base_fame = None
    pick_and_place = None
    pick_index = None
    end_effector_link_name = None

    def ToString(self):
        return str(self)

    def __str__(self):
        return "Planning result for element {} with {} points".format(self.requested_element_id, len(self.trajectory.points))

    def format_trajectory(self, trajectory):
        configs_dicts = []
        if trajectory:
            for point in trajectory.points:
                # Merge trajectory point with start config to make sure they are all full configurations
                # In the past, this was hardcoded only for the RFL, but it makes sense to do it for all
                point = trajectory.start_configuration.merged(point)
                joints_dict = point.joint_dict
                configs_dicts.append(joints_dict)
        return configs_dicts



import json
import os
import sys
import threading

import clr
from compas.data import json_dumps

from compas_xr.realtime_database.realtime_database_interface import RealtimeDatabaseInterface

try:
    from urllib.request import urlopen
except ImportError:
    from urllib import urlopen


lib_dir = os.path.join(os.path.dirname(__file__), "..", "dependencies")
if lib_dir not in sys.path:
    sys.path.append(lib_dir)

clr.AddReference("Firebase.Auth.dll")
clr.AddReference("Firebase.dll")
clr.AddReference("LiteDB.dll")
clr.AddReference("System.Reactive.dll")

from Firebase.Database import FirebaseClient  # noqa: E402
from Firebase.Database.Query import QueryExtensions  # noqa: E402


class RealtimeDatabase(RealtimeDatabaseInterface):
    """
    A RealtimeDatabase is defined by a Firebase configuration path and database reference.

    The RealtimeDatabase class is responsible for initializing and managing the connection to a Firebase Realtime Database.
    It ensures that the database connection is established only once and shared across all instances of the class.

    Parameters
    ----------
    config_path : str
        The path to the Firebase configuration JSON file.

    Attributes
    ----------
    config_path : str
        The path to the Firebase configuration JSON file.
    database : FirebaseClient
        The FirebaseClient instance representing the connection to the Firebase Realtime Database.
    _shared_database : FirebaseClient, class attribute
        The shared FirebaseClient instance representing the connection to the Firebase Realtime Database.
    """

    _shared_database = None

    def __init__(self, config_path):
        self.config_path = config_path
        self.database = self._ensure_database()

    def _ensure_database(self):
        """
        Ensures that the database connection is established.
        If the connection is not yet established, it initializes it.
        If the connection is already established, it returns the existing connection.
        """
        if not RealtimeDatabase._shared_database:
            path = self.config_path
            if not os.path.exists(path):
                raise Exception("Could not find config file at path {}!".format(path))
            with open(path) as config_file:
                config = json.load(config_file)
            # TODO: Database Authorization (Works only with public databases)
            database_url = config["databaseURL"]
            database_client = FirebaseClient(database_url)
            RealtimeDatabase._shared_database = database_client

        if not RealtimeDatabase._shared_database:
            raise Exception("Could not initialize Database!")

        return RealtimeDatabase._shared_database

    def _start_async_call(self, fn, timeout=10):
        """
        Manages asynchronous calls to the RealtimeDatabase.
        """
        result = {}
        result["event"] = threading.Event()
        async_thread = threading.Thread(target=fn, args=(result,))
        async_thread.start()
        async_thread.join(timeout=timeout)
        return result["data"]

    def _get_file_from_remote(self, url):
        """
        This function is used to get the information form the source url and returns a string
        It also checks if the data is None or == null (firebase return if no data)
        """
        try:
            file_content = urlopen(url).read()
        except Exception as e:
            raise Exception("Unable to get file from url {}. Error={}".format(url, str(e)))

        if file_content is not None and file_content != "null":
            return file_content
        else:
            raise Exception("unable to get file from url {}".format(url))

    # TODO: Can this be configured to be a global callback for all methods where I pass the task and the result? This would simplify the code a lot.
    def _task_callback(self, task, result):
        task_awaiter = task.GetAwaiter()
        task_awaiter.OnCompleted(lambda: result["event"].set())
        result["event"].wait()
        result["data"] = True

    def construct_reference(self, parentname):
        """
        Constructs a database reference under the specified parent name.

        Parameters
        ----------
        parentname : str
            The name of the parent under which the reference will be constructed.

        Returns
        -------
        :class: 'Firebase.Database.Query.ChildQuery'
            The constructed database reference.

        """
        database_reference = RealtimeDatabase._shared_database
        reference = database_reference.Child(parentname)
        return reference

    def construct_child_refrence(self, parentname, childname):
        """
        Constructs a database reference under the specified parent name & child name.

        Parameters
        ----------
        parentname : str
            The name of the parent under which the reference will be constructed.
        childname : str
            The name of the child under which the reference will be constructed.

        Returns
        -------
        :class: 'Firebase.Database.Query.ChildQuery'
            The constructed database reference.

        """
        database_reference = RealtimeDatabase._shared_database
        childquery = database_reference.Child(parentname)
        child_reference = QueryExtensions.Child(childquery, childname)
        return child_reference

    def construct_grandchild_refrence(self, parentname, childname, grandchildname):
        """
        Constructs a database reference under the specified parent name, child name, & grandchild name.

        Parameters
        ----------
        parentname : str
            The name of the parent under which the reference will be constructed.
        childname : str
            The name of the child under which the reference will be constructed.
        grandchildname : str
            The name of the grandchild under which the reference will be constructed.

        Returns
        -------
        :class: 'Firebase.Database.Query.ChildQuery'
            The constructed database reference.

        """
        database_reference = RealtimeDatabase._shared_database
        childquery = database_reference.Child(parentname)
        child_reference = QueryExtensions.Child(childquery, childname)
        grand_child_reference = QueryExtensions.Child(child_reference, grandchildname)
        return grand_child_reference

    def construct_reference_from_list(self, reference_list):
        """
        Constructs a database reference under the specified refrences in list order.

        Parameters
        ----------
        reference_list : list of str
            The name of the parent under which the reference will be constructed.

        Returns
        -------
        :class: 'Firebase.Database.Query.ChildQuery'
            The constructed database reference.

        """
        reference = RealtimeDatabase._shared_database
        for ref in reference_list:
            if ref == reference_list[0]:
                reference = reference.Child(ref)
            else:
                reference = QueryExtensions.Child(reference, ref)
        return reference

    def delete_data_from_reference(self, database_reference):
        """
        Method for deleting data from a constructed database reference.

        Parameters
        ----------
        database_reference: 'Firebase.Database.Query.ChildQuery'
            Reference to the database location where the data will be deleted from.

        Returns
        -------
        None
        """
        self._ensure_database()

        def _begin_delete(result):
            deletetask = database_reference.DeleteAsync()
            delete_data = deletetask.GetAwaiter()
            delete_data.OnCompleted(lambda: result["event"].set())
            result["event"].wait()
            result["data"] = True

        self._start_async_call(_begin_delete)

    def get_data_from_reference(self, database_reference):
        """
        Method for retrieving data from a constructed database reference.

        Parameters
        ----------
        database_reference: 'Firebase.Database.Query.ChildQuery'
            Reference to the database location where the data will be retreived from.

        Returns
        -------
        dict
            The retrieved data as a dictionary.

        """
        self._ensure_database()

        def _begin_build_url(result):
            urlbuldtask = database_reference.BuildUrlAsync()
            task_url = urlbuldtask.GetAwaiter()
            task_url.OnCompleted(lambda: result["event"].set())
            result["event"].wait()
            result["data"] = urlbuldtask.Result

        url = self._start_async_call(_begin_build_url)
        json_data = self._get_file_from_remote(url)

        # TODO: json.load(data) vs. json_loads(data)
        """
        This is because error will be thrown with json_loads(data)...
        Because I cannot gaurentee that all data will be filled (FIREBASE DOES NOT UPLOAD NULL VALUES)
        Therefore, unless the data is completely filled json_loads(data) will throw an error.

        """
        data = json.loads(json_data)
        return data

    def stream_data_from_reference(self, callback, database_reference):
        raise NotImplementedError("Function Under Developement")

    def upload_data_to_reference(self, data, database_reference):
        """
        Method for uploading data to a constructed database reference.

        Parameters
        ----------
        data : Any
            The data to be uploaded. Data should be JSON serializable.
        database_reference: 'Firebase.Database.Query.ChildQuery'
            Reference to the database location where the data will be uploaded.

        Returns
        -------
        None
        """
        self._ensure_database()
        serialized_data = json_dumps(data)

        def _begin_upload(result):
            uploadtask = database_reference.PutAsync(serialized_data)
            task_upload = uploadtask.GetAwaiter()
            task_upload.OnCompleted(lambda: result["event"].set())
            result["event"].wait()
            result["data"] = True

        self._start_async_call(_begin_upload)



import io
import json
import os

import pyrebase
from compas.data import json_dumps
from compas.data import json_loads

from compas_xr.storage.storage_interface import StorageInterface

try:
    # from urllib.request import urlopen
    from urllib.request import urlopen
except ImportError:
    from urllib import urlopen


class Storage(StorageInterface):
    """
    A Storage is defined by a Firebase configuration path and a shared storage reference.

    The Storage class is responsible for initializing and managing the connection to a Firebase Storage.
    It ensures that the storage connection is established only once and shared across all instances of the class.

    Parameters
    ----------
    config_path : str
        The path to the Firebase configuration JSON file.

    Attributes
    ----------
    config_path : str
        The path to the Firebase configuration JSON file.
    _shared_storage : pyrebase.Storage, class attribute
        The shared pyrebase.Storage instance representing the connection to the Firebase Storage.
    """

    _shared_storage = None

    def __init__(self, config_path):
        self.config_path = config_path
        self._ensure_storage()

    def _ensure_storage(self):
        """
        Ensures that the storage connection is established.
        If the connection is not yet established, it initializes it.
        If the connection is already established, it returns the existing connection.
        """
        if not Storage._shared_storage:
            path = self.config_path
            if not os.path.exists(path):
                raise Exception("Path Does Not Exist: {}".format(path))
            with open(path) as config_file:
                config = json.load(config_file)
            # TODO: Authorization for storage security (Works for now for us because our Storage is public)
            firebase = pyrebase.initialize_app(config)
            Storage._shared_storage = firebase.storage()

        if not Storage._shared_storage:
            raise Exception("Could not initialize storage!")

    def _get_file_from_remote(self, url):
        """
        This function is used to get the information form the source url and returns a string
        It also checks if the data is None or == null (firebase return if no data)
        """
        try:
            file_content = urlopen(url).read().decode()
        except Exception as e:
            raise Exception("Unable to get file from url {}. Error={}".format(url, str(e)))

        if file_content is not None and file_content != "null":
            return file_content

        else:
            raise Exception("unable to get file from url {}".format(url))

    def construct_reference(self, cloud_file_name):
        """
        Constructs a storage reference for the specified cloud file name.

        Parameters
        ----------
        cloud_file_name : str
            The name of the cloud file.

        Returns
        -------
        :class: 'pyrebase.pyrebase.Storage'
            The constructed storage reference.

        """
        return Storage._shared_storage.child(cloud_file_name)

    def construct_reference_with_folder(self, cloud_folder_name, cloud_file_name):
        """
        Constructs a storage reference for the specified cloud folder name and file name.

        Parameters
        ----------
        cloud_folder_name : str
            The name of the cloud folder.
        cloud_file_name : str
            The name of the cloud file.

        Returns
        -------
        :class: 'pyrebase.pyrebase.Storage'
            The constructed storage reference.

        """
        return Storage._shared_storage.child(cloud_folder_name).child(cloud_file_name)

    def construct_reference_from_list(self, cloud_path_list):
        """
        Constructs a storage reference for consecutive cloud folders in list order.

        Parameters
        ----------
        cloud_path_list : list of str
            The list of cloud path names.

        Returns
        -------
        :class: 'pyrebase.pyrebase.Storage'
            The constructed storage reference.

        """
        storage_reference = Storage._shared_storage
        for path in cloud_path_list:
            storage_reference = storage_reference.child(path)
        return storage_reference

    def get_data_from_reference(self, storage_reference):
        """
        Retrieves data from the specified storage reference.

        Parameters
        ----------
        storage_reference : pyrebase.pyrebase.Storage
            The storage reference pointing to the desired data.

        Returns
        -------
        data : dict or Compas Class Object
            The deserialized data retrieved from the storage reference.

        """
        url = storage_reference.get_url(token=None)
        data = self._get_file_from_remote(url)
        deserialized_data = json_loads(data)
        return deserialized_data

    def upload_bytes_to_reference_from_local_file(self, file_path, storage_reference):
        """
        Uploads data from bytes to the specified storage reference from a local file.

        Parameters
        ----------
        file_path : str
            The path to the local file.
        storage_reference : pyrebase.pyrebase.Storage
            The storage reference to upload the byte data to.

        Returns
        ------
        None

        """
        if not os.path.exists(file_path):
            raise FileNotFoundError("File not found: {}".format(file_path))
        with open(file_path, "rb") as file:
            byte_data = file.read()
        storage_reference.put(byte_data)

    def upload_data_to_reference(self, data, storage_reference, pretty=True):
        """
        Uploads data to the specified storage reference.

        Parameters
        ----------
        data : Any should be json serializable
            The data to be uploaded.
        storage_reference : pyrebase.pyrebase.Storage
            The storage reference to upload the data to.
        pretty : bool, optional
            Whether to format the JSON data with indentation and line breaks (default is True).

        Returns
        ------
        None

        """
        serialized_data = json_dumps(data, pretty=pretty)
        file_object = io.BytesIO(serialized_data.encode())
        storage_reference.put(file_object)



"""
Component to handle execution of trajectory.

Gets trajectories to be executed by a robot.

COMPAS XR v0.9.1
"""

from compas_eve import Subscriber
from compas_eve import Topic
from compas_eve.ghpython import BackgroundWorker
from compas_eve.mqtt import MqttTransport
from ghpythonlib.componentbase import executingcomponent as component

from compas_xr.mqtt import SendTrajectory


def start_server(worker, options):
    topic_name_request = "compas_xr/send_trajectory/" + options.project_name

    worker.count = 0

    def execute_trajectory_requested(request_message):
        worker.count += 1
        worker.display_message("Request #{} started".format(worker.count))
        worker.update_result(request_message, 10)

    tx = MqttTransport(options.host)
    topic = Topic(topic_name_request, SendTrajectory)
    worker.subscriber = Subscriber(topic, callback=execute_trajectory_requested, transport=tx)
    worker.subscriber.subscribe()
    worker.display_message("Subscribed")


def stop_server(worker):
    if hasattr(worker, "subscriber"):
        worker.subscriber.unsubscribe()
    worker.display_message("Stopped")


class ExecuteTrajectoryServiceComponent(component):
    def RunScript(self, options, reset, on):
        if not on:
            BackgroundWorker.stop_instance_by_component(ghenv)  # noqa: F821
            return None

        self.worker = BackgroundWorker.instance_by_component(
            ghenv,  # noqa: F821
            start_server,
            dispose_function=stop_server,
            force_new=reset,
            auto_set_done=False,
            args=(options,),
        )

        if not self.worker.is_working() and not self.worker.is_done() and reset:
            self.worker.start_work()

        if hasattr(self.worker, "result"):
            element_id = self.worker.result.element_id
            robot_name = self.worker.result.robot_name
            return element_id, robot_name
        else:
            return None, None



import os

from compas.datastructures import Assembly
from compas.datastructures import Mesh
from compas.datastructures import Part
from compas.geometry import Frame
from compas.geometry import Point
from compas.geometry import Transformation
from compas.geometry import Vector
from compas_timber.consumers import BrepGeometryConsumer


class AssemblyExtensions(object):
    """
    AssemblyExtensions is a class for extending the functionality of the :class:`~compas.datastructures.Assembly` class.

    The AssemblyExtensions class provides additional functionalities such as exporting parts as .obj files
    and creating a frame assembly from a list of :class:`~compas.geometry.Frame` with a specific data structure
    for localization information.

    """

    def export_timberassembly_objs(self, assembly, folder_path, new_folder_name, z_to_y_remap=False):
        """
        Export timber assembly beams as .obj files to a folder path.

        Parameters
        ----------
        assembly : :class:`~compas_timber.assembly.TimberAssembly`
            The assembly that you want to export beams from.
        folder_path : str
            The path in which you would like to create a storage folder.
        new_folder_name : str
            The name of the folder you would like to create.
        z_to_y_remap : bool, optional
            A boolean that determines if the z-axis should be remapped to the y-axis for .obj export. Default is False.

        Returns
        -------
        None

        """
        target_folder_path = os.path.join(folder_path, new_folder_name)
        if not os.path.exists(target_folder_path):
            os.makedirs(target_folder_path)

        if z_to_y_remap:
            frame = Frame(Point(0, 0, 0), Vector.Xaxis(), Vector.Zaxis())
        else:
            frame = Frame.worldXY()

        for result in BrepGeometryConsumer(assembly).result:
            brep = result.geometry
            beam_frame = result.beam.frame
            key = result.beam.key

            brep_meshes = brep.to_meshes()
            compas_mesh = Mesh()
            for mesh in brep_meshes:
                compas_mesh.join(mesh)
            mesh_transformed = compas_mesh.transformed(Transformation.from_frame_to_frame(beam_frame, frame))

            filename = "{}.obj".format(str(key))
            mesh_transformed.to_obj(os.path.join(target_folder_path, filename))

    def export_mesh_assembly_objs(self, assembly, folder_path, new_folder_name, z_to_y_remap=False):
        """
        Export Mesh assembly parts as .obj files to a folder path.

        Parameters
        ----------
        assembly : :class:`~compas.datastructures.Assembly`
            The Mesh assembly that you want to export parts from.
        folder_path : str
            The path in which you would like to create a storage folder.
        new_folder_name : str
            The name of the folder you would like to create.
        z_to_y_remap : bool, optional
            A boolean that determines if the z-axis should be remapped to the y-axis for .obj export. Default is False.

        Returns
        -------
        None

        """
        target_folder_path = os.path.join(folder_path, new_folder_name)
        if not os.path.exists(target_folder_path):
            os.makedirs(target_folder_path)

        if z_to_y_remap:
            frame = Frame(Point(0, 0, 0), Vector.Xaxis(), Vector.Zaxis())
        else:
            frame = Frame.worldXY()

        for part in assembly.parts():
            # Mesh assembly can be made with or without a frame (ex. assembly.add_part(Mesh)) try & default to worldXY
            if hasattr(part, "frame"):
                part_frame = part.frame
            else:
                part_frame = Frame.worldXY()

            # TODO: This is weird, but I can't transform a Part object, so I need to check if it's a Part or a Mesh
            transformation = Transformation.from_frame_to_frame(part_frame, frame)
            if isinstance(part, Part):
                part_transformed = part.attributes["shape"].transformed(transformation)
            else:
                part_transformed = part.transformed(transformation)

            filename = "{}.obj".format(str(part.key))
            part_transformed.to_obj(os.path.join(target_folder_path, filename))

    def create_qr_assembly(self, qr_frames):
        """
        Create a frame assembly from a list of :class:`~compas.geometry.Frame` with a specific data structure for localization.

        Parameters
        ----------
        qr_frames : list of :class:`~compas.geometry.Frame`
            A list of frames at specific locations for localization data.

        Returns
        -------
        :class:`~compas.datastructures.Assembly`
            The constructed database reference.

        """
        assembly = Assembly()
        for i, frame in enumerate(qr_frames):
            name = "QR_{}".format(i)
            part = Part(name=name, frame=frame, shape=frame)
            assembly.add_part(part)
        return assembly



"""
********************************************************************************
compas_xr.storage
********************************************************************************

This package contains classes for data storage using Firebase.

.. currentmodule:: compas_xr.storage

Classes
-------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    Storage

"""

import sys

if sys.platform == "cli":
    from compas_xr.storage.storage_cli import Storage
else:
    from compas_xr.storage.storage_pyrebase import Storage

__all__ = ["Storage"]



"""
Planning service response.

COMPAS XR v0.9.1
"""

from compas_eve import Publisher
from compas_eve import Topic
from compas_eve.mqtt import MqttTransport
from ghpythonlib.componentbase import executingcomponent as component

from compas_xr.mqtt import GetTrajectoryResult


class PlanningServiceResponseComponent(component):
    def RunScript(self, options, result, publish):
        if not result:
            self.Message = "Null Result, unable to publish"
            return

        if publish:
            topic_name_result = "compas_xr/get_trajectory_result/" + options.project_name
            topic = Topic(topic_name_result, GetTrajectoryResult)
            tx = MqttTransport(options.host)
            publisher = Publisher(topic, transport=tx)
            message = GetTrajectoryResult(
                element_id=result.requested_element_id,
                robot_name=options.robot_name,
                robot_base_frame=result.robot_base_frame,
                trajectory=result.trajectory,
                pick_and_place=result.pick_and_place,
                pick_index=result.pick_index,
                end_effector_link_name=result.end_effector_link_name,
            )
            publisher.publish(message)
            self.Message = "Send trajectory for #{}".format(result.requested_element_id)
